"""Wrapper around Yahoo Finance API
https://query1.finance.yahoo.com/v7/finance/download/aapl?period1=1487685205&period2=1495457605&interval=1d&events=history&crumb=dU2hYSfAy9E
https://query1.finance.yahoo.com/v7/finance/quote?symbols=TLS.AX,MUS.AX
https://help.yahoo.com/kb/SLN2310.html
"""
import csv
import datetime
import logging
import re
from collections import namedtuple
from decimal import Decimal
from time import sleep

import pytz


MAX_ATTEMPT = 5

YAHOO_BASE_URL = "https://query1.finance.yahoo.com/v7/finance"

YahooSymbol = namedtuple("YahooSymbol", "name,symbol,exchange,timezone,currency,date,price")
YahooQuote = namedtuple("YahooQuote", "date,open,high,low,close,adj_close,volume")


def get_latest_quote(symbol):
    import requests

    resp = requests.get("{}/quote".format(YAHOO_BASE_URL), params={"symbols": symbol})
    resp.raise_for_status()

    try:
        data = resp.json()["quoteResponse"]["result"][0]
    except IndexError:
        from .core.commodity import GncCommodityError

        raise GncCommodityError("Can't find information on symbol '{}' on yahoo".format(symbol))

    tz = data["exchangeTimezoneName"]

    return YahooSymbol(
        data["longName"],
        data["symbol"],
        data["exchange"],
        tz,
        data["currency"],
        datetime.datetime.fromtimestamp(data["regularMarketTime"]).astimezone(pytz.timezone(tz)),
        data["regularMarketPrice"],
    )


crumble_link = "https://finance.yahoo.com/quote/{0}/history?p={0}"
crumble_regex = r'CrumbStore":{"crumb":"(.*?)"}'
quote_link = "https://query1.finance.yahoo.com/v7/finance/download/{}?period1={}&period2={}&interval=1d&events=history&crumb={}"


def get_crumble_and_cookie(symbol):
    import requests

    link = crumble_link.format(symbol)
    response = requests.get(link)
    cookie_str = response.headers["set-cookie"]

    match = re.search(crumble_regex, response.text)
    crumble_str = match.group(1)
    return crumble_str, cookie_str


def download_quote(symbol, date_from, date_to, tz=None):
    import requests

    def normalize(d):
        if isinstance(d, datetime.datetime):
            pass
        elif isinstance(d, datetime.date):
            d = datetime.datetime.combine(d, datetime.time(0))
        else:
            d = datetime.datetime.strptime(d, "%Y-%m-%d")
        if not d.tzinfo:
            assert tz
            # todo: understand yahoo behavior as even in the browser, I get
            # weird results ...
            # d = d.replace(tzinfo=tz)
        return d

    date_from = normalize(date_from)
    date_to = normalize(date_to)
    time_stamp_from = int(date_from.timestamp())
    time_stamp_to = int(date_to.timestamp())

    for i in range(MAX_ATTEMPT):
        logging.info("{} attempt to download quotes for symbol {} from yahoo".format(i, symbol))

        crumble_str, cookie_str = get_crumble_and_cookie(symbol)

        link = quote_link.format(symbol, time_stamp_from, time_stamp_to, crumble_str)

        resp = requests.get(link, headers={"Cookie": cookie_str})
        try:
            resp.raise_for_status()
        except Exception as e:
            sleep(2)
        else:
            break
    else:
        raise e

    csv_data = list(csv.reader(resp.text.strip().split("\n")))

    return [
        yq
        for data in csv_data[1:]
        for yq in [
            YahooQuote(
                datetime.datetime.strptime(data[0], "%Y-%m-%d").date(),
                *[Decimal(f) for f in data[1:]]
            )
        ]
        if date_from.date() <= yq.date <= date_to.date()
    ]


if __name__ == "__main__":
    print(get_latest_quote("KO"))

    print(download_quote("ENGI.PA", "2018-02-26", "2018-03-01", tz=pytz.timezone("CET")))

if __name__ == "__main__":
    print(get_latest_quote("ENGI.PA"))
    print(get_latest_quote("AAPL"))



from __future__ import print_function
import os
import tempfile

from piecash import open_book, create_book, GnucashException


FILE_1 = os.path.join(tempfile.gettempdir(), "not_there.gnucash")
FILE_2 = os.path.join(tempfile.gettempdir(), "example_file.gnucash")

if os.path.exists(FILE_2):
    os.remove(FILE_2)

# open a file that isn't there, detect the error
try:
    book = open_book(FILE_1)
except GnucashException as backend_exception:
    print("OK", backend_exception)

# create a new file, this requires a file type specification
with create_book(FILE_2) as book:
    pass

# open the new file, try to open it a second time, detect the lock
# using the session as context manager automatically release the lock and close the session
with open_book(FILE_2) as book:
    try:
        with open_book(FILE_2) as book_2:
            pass
    except GnucashException as backend_exception:
        print("OK", backend_exception)

os.remove(FILE_2)



"""
Read currency exchange rates.
This example is similar to read_currencies script but uses SQLAlchemy directly
to achieve the same result.

The goal is simply to illustrate what happens under the hood. Hopefully this 
shows how simple the piecash API is.
"""
# pylint: disable=invalid-name
import sys
import piecash
from piecash import Commodity

# Variables
filename = sys.argv[1]
if filename is None:
    print("You need to specify a valid .gnucash file to use.")
    filename = "test.gnucash"

symbol = "AUD"
####################################

with piecash.open_book(filename, open_if_lock=True) as book:
    # , readonly=False,

    # SQLAlchemy session.
    session = book.session

    # query example:
    #accountsFiltered = session.query(Account).filter(Account.name >= "T").all()
    # SQLAlchemy methods: count, first, all, one...

    # Get all the currencies in the book (i.e. for update).
    print("All currencies used in the book:")
    currencies = session.query(Commodity).filter(Commodity.namespace == "CURRENCY").all()
    for c in currencies:
        print(c)

    # Accessing individual records.

    print("\nSelected single currency details (" + symbol + "):")
    cdty = session.query(Commodity).filter(Commodity.namespace == "CURRENCY", Commodity.mnemonic == symbol).first()

    # Accessing attributes of a commodity.
    print("Commodity namespace={cdty.namespace}\n"
        "          mnemonic={cdty.mnemonic}\n"
        "          cusip={cdty.cusip}\n"
        "          fraction={cdty.fraction}".format(cdty=cdty))

    # Loop through the existing commodity prices.
    # This can be used to fetch the points for a price graph.
    print("\nHistorical prices:")
    for pr in cdty.prices:
        print("Price date={pr.date}"
            "      value={pr.value} {pr.currency.mnemonic}/{pr.commodity.mnemonic}".format(pr=pr))



from piecash import create_book, Account

# create a book with some account tree structure
with create_book("../gnucash_books/simple_book_transaction_creation.gnucash", overwrite=True) as mybook:
    mybook.root_account.children = [
        Account(name="Expenses",
                type="EXPENSE",
                commodity=mybook.currencies(mnemonic="USD"),
                placeholder=True,
                children=[
                    Account(name="Some Expense Account",
                            type="EXPENSE",
                            commodity=mybook.currencies(mnemonic="USD")),
                ]),
        Account(name="Assets",
                type="ASSET",
                commodity=mybook.currencies(mnemonic="USD"),
                placeholder=True,
                children=[
                    Account(name="Current Assets",
                            type="BANK",
                            commodity=mybook.currencies(mnemonic="USD"),
                            placeholder=True,
                            children=[
                                Account(name="Checking",
                                        type="BANK",
                                        commodity=mybook.currencies(mnemonic="USD"))
                            ]),
                ]),
    ]
    # save the book
    mybook.save()

from piecash import open_book, Transaction, Split
from datetime import datetime
from decimal import Decimal

# reopen the book and add a transaction
with open_book("../gnucash_books/simple_book_transaction_creation.gnucash",
               open_if_lock=True,
               readonly=False) as mybook:
    today = datetime.now()
    # retrieve the currency from the book
    USD = mybook.currencies(mnemonic="USD")
    # define the amount as Decimal
    amount = Decimal("25.35")
    # retrieve accounts
    to_account = mybook.accounts(fullname="Expenses:Some Expense Account")
    from_account = mybook.accounts(fullname="Assets:Current Assets:Checking")
    # create the transaction with its two splits
    Transaction(
        post_date=today.date(),
        enter_date=today,
        currency=USD,
        description="Transaction Description!",
        splits=[
            Split(account=to_account,
                  value=amount,
                  memo="Split Memo!"),
            Split(account=from_account,
                  value=-amount,
                  memo="Other Split Memo!"),
        ]
    )
    # save the book
    mybook.save()

from piecash import ledger

# check the book by exporting to ledger format
with open_book("../gnucash_books/simple_book_transaction_creation.gnucash",
               open_if_lock=True) as mybook:
    print(ledger(mybook))


# coding=utf-8
from __future__ import unicode_literals

from collections import defaultdict
from datetime import datetime, date, time
from decimal import Decimal

import pytest

from piecash import Transaction, Split, GncImbalanceError, GncValidationError, Lot
from test_helper import db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_basic, book_transactions

# dummy line to avoid removing unused symbols

a = db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_basic, book_transactions


class TestTransaction_create_transaction(object):
    def test_create_basictransaction_neutraltime(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        e = book_basic.accounts(name="exp")

        tr = Transaction(currency=EUR, description=u"wire from Hélène", notes=u"on St-Eugène day",
                         post_date=date(2014, 1, 1),
                         splits=[
                             Split(account=a, value=100, memo=u"mémo asset"),
                             Split(account=e, value=-100, memo=u"mémo exp"),
                         ])

        assert isinstance(tr.post_date, date)

        book_basic.flush()
        book_basic.validate()

        assert isinstance(tr.post_date, date)

    def test_create_basictransaction_validation_date(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        e = book_basic.accounts(name="exp")

        splits = [
            Split(account=a, value=100, memo=u"mémo asset"),
            Split(account=e, value=-10, memo=u"mémo exp"),
        ]

        with pytest.raises(GncValidationError):
            tr = Transaction(currency=EUR, description=u"wire from Hélène", notes=u"on St-Eugène day",
                             post_date=datetime(2014, 1, 1),
                             enter_date=datetime(2014, 1, 1),
                             splits=splits)

        with pytest.raises(GncValidationError):
            tr = Transaction(currency=EUR, description=u"wire from Hélène", notes=u"on St-Eugène day",
                             post_date=datetime(2014, 1, 1),
                             enter_date=time(10, 59, 00),
                             splits=splits)

        with pytest.raises(GncValidationError):
            tr = Transaction(currency=EUR, description=u"wire from Hélène", notes=u"on St-Eugène day",
                             post_date=date(2014, 1, 1),
                             enter_date=date(2014, 1, 1),
                             splits=splits)

        tr = Transaction(currency=EUR, description=u"wire from Hélène", notes=u"on St-Eugène day",
                         post_date=None,
                         enter_date=None,
                         splits=splits)

        with pytest.raises(GncImbalanceError):
            book_basic.flush()
            book_basic.validate()

    def test_create_basictransaction(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        e = book_basic.accounts(name="exp")

        tr = Transaction(currency=EUR, description=u"wire from Hélène", notes=u"on St-Eugène day",
                         post_date=date(2014, 1, 1),
                         enter_date=datetime(2014, 1, 1),
                         splits=[
                             Split(account=a, value=100, memo=u"mémo asset"),
                             Split(account=e, value=-10, memo=u"mémo exp"),
                         ])
        # check issue with balance
        with pytest.raises(GncImbalanceError):
            book_basic.flush()
            book_basic.validate()

        # adjust balance
        Split(account=e, value=-90, memo="missing exp", transaction=tr)
        book_basic.flush()

        # check no issue with str
        assert str(tr)
        assert str(tr.splits)
        assert repr(tr)
        assert repr(tr.splits)
        assert tr.notes == "on St-Eugène day"

    def test_create_basictransaction_splitfirst(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        e = book_basic.accounts(name="exp")
        s = Split(account=a, value=Decimal(1))
        assert repr(s)

    def test_create_cdtytransaction(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        s = book_basic.accounts(name="broker")

        tr = Transaction(currency=EUR, description="buy stock", notes=u"on St-Eugène day",
                         post_date=date(2014, 1, 2),
                         enter_date=datetime(2014, 1, 3),
                         splits=[
                             Split(account=a, value=100, memo=u"mémo asset"),
                             Split(account=s, value=-90, memo=u"mémo brok"),
                         ])

        # check issue with quantity for broker split not defined
        with pytest.raises(GncValidationError):
            book_basic.validate()

        sb = tr.splits(account=s)
        sb.quantity = 15

        # check issue with quantity not same sign as value
        with pytest.raises(GncValidationError):
            book_basic.validate()

        sb.quantity = -15

        # verify imbalance issue
        with pytest.raises(GncImbalanceError):
            book_basic.validate()

        # adjust balance
        Split(account=a, value=-10, memo="missing asset corr", transaction=tr)
        book_basic.save()
        assert str(sb)
        assert str(sb)

        # changing currency of an existing transaction is not allowed
        tr.currency = book_basic.currencies(mnemonic="USD")
        with pytest.raises(GncValidationError):
            book_basic.validate()
        book_basic.cancel()

        # check sum of quantities are not balanced per commodity but values are
        d = defaultdict(lambda: Decimal(0))
        for sp in tr.splits:
            assert sp.quantity == sp.value or sp.account != a
            d[sp.account.commodity] += sp.quantity
            d["cur"] += sp.value
        assert d["cur"] == 0
        assert all([v != 0 for k, v in d.items() if k != "cur"])

    def test_create_split_overflow(self, book_basic):
        a = book_basic.accounts(name="asset")

        # raise error as Transaction has a non CURRENCY commodity
        with pytest.raises(TypeError):
            sp = Split(account=a, value=1. / 3., quantity=10, memo=u"mémo asset")

        with pytest.raises(ValueError):
            sp = Split(account=a, value=Decimal(1) / Decimal(3), quantity=10, memo=u"mémo asset")

        sp = Split(account=a, value=Decimal(1234567890123455678), quantity=10, memo=u"mémo asset")

        with pytest.raises(ValueError):
            sp = Split(account=a, value=Decimal(1234567890123455678901234), quantity=10, memo=u"mémo asset")

    def test_create_cdtytransaction_cdtycurrency(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        s = book_basic.accounts(name="broker")

        tr = Transaction(currency=s.commodity, description="buy stock", notes=u"on St-Eugène day",
                         post_date=date(2014, 1, 2),
                         enter_date=datetime(2014, 1, 3),
                         splits=[
                             Split(account=a, value=100, quantity=10, memo=u"mémo asset"),
                             Split(account=s, value=-100, quantity=-10, memo=u"mémo brok"),
                         ])
        # raise error as Transaction has a non CURRENCY commodity
        with pytest.raises(GncValidationError):
            book_basic.validate()

    def test_create_cdtytransaction_tradingaccount(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        s = book_basic.accounts(name="broker")

        book_basic.use_trading_accounts = True
        tr = Transaction(currency=EUR, description="buy stock", notes=u"on St-Eugène day",
                         post_date=date(2014, 1, 2),
                         enter_date=datetime(2014, 1, 3),
                         splits=[
                             Split(account=a, value=100, memo=u"mémo asset"),
                             Split(account=s, value=-100, quantity=-15, memo=u"mémo brok"),
                         ])
        book_basic.validate()

        assert "{}".format(tr) == "Transaction<[EUR] 'buy stock' on 2014-01-02>"
        assert "{}".format(s) == "Account<asset:broker[ïoà]>"
        assert "{}".format(tr.splits(account=s)) == "Split<Account<asset:broker[ïoà]> -100 EUR [-15 ïoà]>"
        assert "{}".format(tr.splits(account=a)) == "Split<Account<asset[EUR]> 100 EUR>"

        # check sum of quantities are all balanced per commodity as values are
        d = defaultdict(lambda: Decimal(0))
        for sp in tr.splits:
            assert sp.quantity == sp.value or sp.account != a
            d[sp.account.commodity] += sp.quantity
            d["cur"] += sp.value

        assert d["cur"] == 0
        assert all([v == Decimal(0) for k, v in d.items() if k != "cur"])

        # change existing quantity
        sp = tr.splits(memo=u"mémo brok")
        sp.quantity += 1
        book_basic.validate()

        # check sum of quantities are all balanced per commodity as values are
        d = defaultdict(lambda: Decimal(0))
        for sp in tr.splits:
            assert sp.quantity == sp.value or sp.account != a
            d[sp.account.commodity] += sp.quantity
            d["cur"] += sp.value
        assert d["cur"] == 0
        assert all([v == Decimal(0) for k, v in d.items() if k != "cur"])

    def test_tag_split_zero_quantity(self, book_transactions):
        broker = book_transactions.accounts(name="broker")
        asset = book_transactions.accounts.get(name="asset")
        inc = book_transactions.accounts.get(name="inc")
        currency = book_transactions.default_currency

        value = Decimal(250)
        splits = [
            Split(asset, value),
            Split(inc, -value),
            Split(broker, value=0, quantity=0),  # tag split for assigning dividend income to stock
        ]

        Transaction(currency, description='Dividend income', splits=splits)

        book_transactions.validate()

    def test_tag_split_zero_quantity_with_value(self, book_transactions):
        broker = book_transactions.accounts(name="broker")
        inc = book_transactions.accounts.get(name="inc")
        value = Decimal(250)

        # Transaction recording capital gains.
        splits = [
            Split(broker, value, quantity=0),
            Split(inc, -value),
        ]
        Transaction(inc.commodity, description='Capital gains', splits=splits)
        book_transactions.validate()

        # Transaction recording capital loss.
        splits = [
            Split(broker, -value, quantity=0),
            Split(inc, value),
        ]
        Transaction(inc.commodity, description='Capital loss', splits=splits)
        book_transactions.validate()

        # Do the same tests with a -0.0 quantity. This Decimal has is_signed=True.
        mzero = Decimal('-0.00')

        # Transaction recording capital gains.
        splits = [
            Split(broker, value, quantity=mzero),
            Split(inc, -value),
        ]
        Transaction(inc.commodity, description='Capital gains', splits=splits)
        book_transactions.validate()

        # Transaction recording capital loss.
        splits = [
            Split(broker, -value, quantity=mzero),
            Split(inc, value),
        ]
        Transaction(inc.commodity, description='Capital loss', splits=splits)
        book_transactions.validate()

    def test_tag_split_zero_value(self, book_transactions):
        broker = book_transactions.accounts(name="broker")
        asset = book_transactions.accounts.get(name="asset")
        currency = book_transactions.default_currency

        # Give away 250 shares for free.
        quantity = Decimal(-250)
        splits = [
            Split(asset, 0),
            Split(broker, 0, quantity=quantity),
        ]

        Transaction(currency, description='donation', splits=splits)

        book_transactions.validate()



class TestTransaction_lots(object):
    def test_create_simpletlot_addsplits(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        s = book_basic.accounts(name="broker")
        l = Lot(title=u"test mé", account=s, notes=u"ïlya")
        for i, am in enumerate([45, -35, -20]):
            tr = Transaction(currency=EUR, description="trade stock", notes=u"àçö",
                             post_date=date(2014, 1, 1 + i),
                             enter_date=datetime(2014, 1, 1 + i),
                             splits=[
                                 Split(account=a, value=am * 10, memo=u"mémo asset"),
                                 Split(account=s, value=-am * 10, quantity=-am, memo=u"mémo brok", lot=l),
                             ])
        book_basic.flush()

    def test_create_simpletlot_initialsplits(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        s = book_basic.accounts(name="broker")
        sp = []
        for i, am in enumerate([45, -35, -20]):
            tr = Transaction(currency=EUR, description="trade stock", notes=u"àçö",
                             post_date=date(2014, 1, 1 + i),
                             enter_date=datetime(2014, 1, 1 + i),
                             splits=[
                                 Split(account=a, value=am * 10, memo=u"mémo asset"),
                                 Split(account=s, value=-am * 10, quantity=-am, memo=u"mémo brok"),
                             ])
            sp.append(tr.splits(account=s))

        l = Lot(title=u"test mé", account=s, notes=u"ïlya", splits=sp)
        book_basic.flush()

    def test_create_closedlot_addsplits(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        s = book_basic.accounts(name="broker")
        l = Lot(title=u"test mé", account=s, notes=u"ïlya")
        l.is_closed = 1
        # raise valueerror as lot is closed
        with pytest.raises(ValueError):
            tr = Transaction(currency=EUR, description="trade stock", notes=u"àçö",
                             post_date=date(2014, 1, 1),
                             enter_date=datetime(2014, 1, 1),
                             splits=[
                                 Split(account=a, value=10, memo=u"mémo asset"),
                                 Split(account=s, value=- 10, quantity=-2, memo=u"mémo brok", lot=l),
                             ])

    def test_create_simplelot_inconsistentaccounts(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        s = book_basic.accounts(name="broker")
        l = Lot(title=u"test mé", account=a, notes=u"ïlya")
        # raise valueerror as split account not the same as lot account
        tr = Transaction(currency=EUR, description="trade stock", notes=u"àçö",
                         post_date=date(2014, 1, 1),
                         enter_date=datetime(2014, 1, 1),
                         splits=[
                             Split(account=a, value=10, memo=u"mémo asset"),
                             Split(account=s, value=- 10, quantity=-2, memo=u"mémo brok", lot=l),
                         ])

        with pytest.raises(ValueError):
            book_basic.validate()


class TestTransaction_changes(object):
    def test_delete_existing_transaction(self, book_transactions):
        l = len(book_transactions.transactions)
        s = len(book_transactions.splits)
        tr = book_transactions.transactions(description="my revenue")
        book_transactions.delete(tr)
        book_transactions.save()
        nl = len(book_transactions.transactions)
        ns = len(book_transactions.splits)
        assert nl == l - 1
        assert ns == s - 2

    def test_change_cdty_split_price(self, book_transactions):
        tr = book_transactions.transactions(description="my purchase of stock")
        sp = tr.splits(account=book_transactions.accounts(name="broker"))

        assert len(book_transactions.prices) == 6

        p = [p for p in book_transactions.prices if p.date.day == 29][0]
        p_expected = (sp.value / sp.quantity).quantize(Decimal("0.000001"))
        assert p.value == p_expected

        # changing the quantity of the split should NOT change the existing price
        sp.quantity = (5, 1)

        book_transactions.validate()

        p_not_expected = (sp.value / sp.quantity).quantize(Decimal("0.000001"))

        assert len(book_transactions.prices) == 6
        assert p.value == p_expected
        assert p_expected != p_not_expected

        # changing the post date of the transaction of the split should create a new price
        tr.post_date = date(2015, 1, 29)
        book_transactions.validate()
        book_transactions.flush()
        assert len(book_transactions.prices) == 7


class TestSplit_credit_debit(object):
    def test_delete_existing_transaction(self, book_transactions):
        asset = book_transactions.accounts.get(name="asset")
        inc = book_transactions.accounts.get(name="inc")
        exp = book_transactions.accounts.get(name="exp")

        # revenue item
        revenue = book_transactions.transactions(description="my revenue")
        revenue_split_asset = revenue.splits(account=asset)
        revenue_split_inc = revenue.splits(account=inc)
        assert revenue_split_asset.is_credit is False
        assert revenue_split_asset.is_debit is True
        assert revenue_split_inc.is_credit is True
        assert revenue_split_inc.is_debit is False

        # expense item
        expense = book_transactions.transactions(description="my expense")
        expense_split_asset = expense.splits(account=asset)
        expense_split_inc = expense.splits(account=exp)
        assert expense_split_asset.is_credit is True
        assert expense_split_asset.is_debit is False
        assert expense_split_inc.is_credit is False
        assert expense_split_inc.is_debit is True



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function

import datetime
import os
import shutil
import threading
from decimal import Decimal
from pathlib import Path

import pytest
import sqlalchemy

from piecash import create_book, Account, ACCOUNT_TYPES, open_book, Price
from piecash._common import GnucashException
from piecash.core.account import _is_parent_child_types_consistent, root_types
from piecash.kvp import Slot
from test_helper import (
    file_template_full,
    file_for_test_full,
    run_file,
    file_ghost_kvp_scheduled_transaction,
    file_ghost_kvp_scheduled_transaction_for_test,
)


@pytest.fixture
def book(request):
    b = create_book()
    return b


@pytest.fixture
def realbook_session(request):
    return use_copied_book(request, file_template_full, file_for_test_full)


@pytest.fixture
def realbook_session_multithread(request):
    return use_copied_book(request, file_template_full, file_for_test_full, check_same_thread=False)


@pytest.fixture
def ghost_kvp_scheduled_transaction_session(request):
    return use_copied_book(request, file_ghost_kvp_scheduled_transaction, file_ghost_kvp_scheduled_transaction_for_test)


def use_copied_book(request, template_filename, test_filename, check_same_thread=True):
    shutil.copy(str(template_filename), str(test_filename))

    # default book is readonly
    s = open_book(test_filename, check_same_thread=check_same_thread)

    @request.addfinalizer
    def finalizer():
        s.close()
        test_filename.unlink()

    return s


class TestIntegration_ExampleScripts(object):
    def test_simple_move_split(self):
        run_file("examples/simple_move_split.py")

    def test_simple_book(self):
        run_file("examples/simple_book.py")

    def test_filtered_transaction_report(self):
        run_file("examples/filtered_transaction_report.py")

    def test_simple_session(self):
        run_file("examples/simple_session.py")

    def test_simple_test(self):
        run_file("examples/simple_test.py")

    def test_simple_sqlite_create(self):
        run_file("examples/simple_sqlite_create.py")


class TestIntegration_EmptyBook(object):
    def test_slots_create_access(self, book):
        kv = {
            "vint": 3,
            "vfl": 2.34,
            "vstr": "hello",
            "vdate": datetime.datetime.now().date(),
            "vtime": datetime.datetime.now(),
            "vnum": Decimal("4.53"),
            "vlist": ["stri", 4, dict(foo=23)],
            "vdct": {"spl": 2.3, "vfr": {"vfr2": {"foo": 33, "baz": "hello"}, "coo": Decimal("4.53")}},
        }
        for k, v in kv.items():
            book[k] = v
        book.save()

        for k, v in kv.items():
            assert k in book
            if isinstance(v, datetime.datetime):
                # check string format as the date in piecash is localized
                assert "{:%Y%m%d%H%M%S}".format(book[k].value) == "{:%Y%m%d%H%M%S}".format(v)
            else:
                assert book[k].value == v

    def test_slots_strings_access(self, book):
        b = book

        b["a/b/c/d/e"] = 1
        book.book.flush()
        assert b["a"]["b"]["c"]["d"]["e"].value == 1

        b["a/b/c"] = {"d": {"t": "ok"}}

        b["a/b/c/d/f"] = "2"
        book.book.flush()
        assert len(b["a"]["b"]["c"]["d"].slots) == 2

        b["a/b/c/d/f"] = "5"
        assert b["a"]["b/c"]["d"]["f"].value == "5"

        for k, v in b["a/b/c/d"].iteritems():
            assert k == "f" or k == "t"
        print(b.slots)
        assert b["a/b/c/d"].get("t", "hello") == "ok"
        assert b["a/b/c/d"].get("not there", "hello") == "hello"

        del b["a/b/c/d/t"]
        assert repr(b["a"]) == "<SlotFrame a={'b': {'c': {'d': {'f': '5'}}}}>"

        with pytest.raises(TypeError):
            b["a/b/c/d/f"] = 4
        with pytest.raises(TypeError):
            b["a/b/c"] = True

        book.flush()
        assert {n for (n,) in book.session.query(Slot._name)} == {"a", "a/b", "a/b/c", "a/b/c/d", "a/b/c/d/f"}

        # delete some elements
        del b["a"]["b"][:]
        book.flush()
        assert {n for (n,) in book.session.query(Slot._name)} == {"a", "a/b"}

        book.flush()
        assert len(b["a"].slots) == 1
        assert len(b["a/b"].slots) == 0

        with pytest.raises(KeyError):
            b["a/b/c"]

        del b["a"]["b"]
        book.session.flush()
        assert len(b["a"].slots) == 0

        with pytest.raises(TypeError):
            b["a"] = b

        with pytest.raises(KeyError):
            del b["a/n"]

        del b[:]
        book.session.flush()
        assert {n for (n,) in book.session.query(Slot._name)} == set([])

    def test_smart_slots(self, book):
        book["account"] = book.root_account
        assert book.slots[0].guid_val == book.root_account.guid
        assert book["account"].value == book.root_account

        with pytest.raises(ValueError):
            book["weird"] = lambda x: x

        with pytest.raises(ValueError):
            book["unknown_guid"] = book.root_account

    def test_empty_gnucash_file(self, book):
        accs = book.accounts

        assert len(accs) == 0
        assert all(acc.parent is None for acc in accs)
        assert all(acc.type == "ROOT" for acc in accs)

    def test_is_parent_child_types_consistent(self):
        combi_OK = [("ROOT", "BANK"), (None, "ROOT"), ("ROOT", "EQUITY"), ("ROOT", "ASSET"), ("ROOT", "EXPENSE")]

        combi_not_OK = [
            ("ROOT", "ROOT"),
            ("ROOT", None),
            (None, "ASSET"),
            ("ASSET", "EQUITY"),
            ("EQUITY", "ASSET"),
            ("ASSET", "INCOME"),
            ("EXPENSE", "ASSET"),
        ]

        for p, c in combi_OK:
            assert _is_parent_child_types_consistent(p, c, [])

        for p, c in combi_not_OK:
            assert not _is_parent_child_types_consistent(p, c, [])

    def test_add_account_compatibility(self, book):
        # test compatibility between child account and parent account
        for acc_type1 in ACCOUNT_TYPES - root_types:
            acc1 = Account(name=acc_type1, type=acc_type1, parent=book.root_account, commodity=None)
            for acc_type2 in ACCOUNT_TYPES:
                if _is_parent_child_types_consistent(acc_type1, acc_type2, []):
                    acc2 = Account(name=acc_type2, type=acc_type2, parent=acc1, commodity=None)

        book.save()

        assert len(book.accounts) == 100

    def test_add_account_incompatible(self, book):
        # test compatibility between child account and parent account
        for acc_type1 in ACCOUNT_TYPES - root_types:
            acc1 = Account(name=acc_type1, type=acc_type1, parent=book.root_account, commodity=None)
        book.save()

        assert len(book.accounts) == 13
        for acc_type1 in ACCOUNT_TYPES - root_types:
            acc1 = book.accounts(name=acc_type1)
            for acc_type2 in ACCOUNT_TYPES:
                if not _is_parent_child_types_consistent(acc_type1, acc_type2, []):
                    acc2 = Account(name=acc_type2, type=acc_type2, parent=acc1, commodity=None)
                    with pytest.raises(ValueError):
                        book.validate()
                    book.cancel()
        book.save()

        assert len(book.accounts) == 13

    def test_add_account_names(self, book):
        # raise ValueError as acc1 and acc2 shares same parents with same name
        acc1 = Account(name="Foo", type="MUTUAL", parent=book.root_account, commodity=None)
        acc2 = Account(name="Foo", type="BANK", parent=book.root_account, commodity=None)
        with pytest.raises(ValueError):
            book.save()
        book.cancel()
        # ok as same name but different parents
        acc3 = Account(name="Fooz", type="BANK", parent=book.root_account, commodity=None)
        acc4 = Account(name="Fooz", type="BANK", parent=acc3, commodity=None)
        book.save()
        # raise ValueError as now acc4 and acc3 shares same parents with same name
        acc4.parent = acc3.parent
        with pytest.raises(ValueError):
            book.save()

    def test_example(self, realbook_session):
        book = realbook_session

        # example 1, print all stock prices in the Book
        # display all prices
        for price in book.query(Price).all():
            print(
                "{}/{} on {} = {} {}".format(
                    price.commodity.namespace,
                    price.commodity.mnemonic,
                    price.date,
                    float(price.value_num) / price.value_denom,
                    price.currency.mnemonic,
                )
            )

        for account in book.accounts:
            print(account)

        # build map between account fullname (e.g. "Assets:Current Assets" and account)
        map_fullname_account = {account.fullname: account for account in book.query(Account).all()}

        # use it to retrieve the current assets account
        print(map_fullname_account)
        acc_cur = map_fullname_account["Assets:Current Assets"]

        # retrieve EUR currency
        EUR = book.commodities.get(mnemonic="EUR")

        # add a new subaccount to this account of type ASSET with currency EUR
        Account(name="new savings account", type="ASSET", parent=acc_cur, commodity=EUR)

        # save changes
        with pytest.raises(GnucashException) as excinfo:
            book.save()


class TestIntegration_GhostKvpScheduledTransaction(object):

    # See PR https://github.com/sdementen/piecash/pull/20
    # The book ghost_kvp_scheduled_transaction.gnucash was created in GnuCash
    # as follows:
    # * Created a scheduled transaction
    # * Used it to create a real transaction
    # * Deleted the original template transaction

    def test_print_transactions(self, ghost_kvp_scheduled_transaction_session):
        book = ghost_kvp_scheduled_transaction_session

        for tr in book.transactions:
            assert tr.scheduled_transaction is None


class TestIntegration_Thread(object):
    def test_exception_due_to_multithread(self, realbook_session):
        book = realbook_session

        class MyThread(threading.Thread):
            def run(self):
                with pytest.raises(sqlalchemy.exc.ProgrammingError):
                    book.transactions

        thr = MyThread()
        thr.start()
        thr.join()

    def test_check_same_thread_False_allow_thread_use(self, realbook_session_multithread):
        book = realbook_session_multithread

        class MyThread(threading.Thread):
            def run(self):
                book.transactions

        thr = MyThread()
        thr.start()
        thr.join()



# coding=utf-8
from __future__ import unicode_literals

from datetime import date
from decimal import Decimal

import pytest

from piecash import Price, Commodity, GnucashException
from piecash.core.commodity import GncPriceError
from test_helper import db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_basic, \
    is_inmemory_sqlite, needweb

# dummy line to avoid removing unused symbols
a = db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_basic


class TestCommodity_create_commodity(object):
    def test_create_commodity(self, book_basic):
        assert len(book_basic.commodities) == 2
        cdty = Commodity(namespace="AMEX", mnemonic="APPLE", fullname="Apple", book=book_basic)
        book_basic.flush()
        assert len(book_basic.commodities) == 3

        with pytest.raises(GnucashException):
            cdty.base_currency

        cdty["quoted_currency"] = "EUR"
        assert cdty.base_currency == book_basic.commodities(mnemonic="EUR")

    def test_create_commodity_uniqueness(self, book_basic):
        assert len(book_basic.commodities) == 2
        cdty1 = Commodity(namespace="AMEX", mnemonic="APPLE", fullname="Apple", book=book_basic)
        cdty2 = Commodity(namespace="AMEX", mnemonic="APPLE", fullname="Apple", book=book_basic)

        with pytest.raises(ValueError):
            book_basic.save()

    def test_base_currency_commodity(self, book_basic):
        cdty = Commodity(namespace="AMEX", mnemonic="APPLE", fullname="Apple", book=book_basic)

        with pytest.raises(GnucashException):
            cdty.base_currency

        # should trigger creation of USD currency
        cdty["quoted_currency"] = "USD"
        assert cdty.base_currency.mnemonic == 'USD'
        book_basic.flush()
        assert cdty.base_currency == book_basic.currencies(mnemonic="USD")

        cdty["quoted_currency"] = "EUR"
        assert cdty.base_currency == book_basic.currencies(mnemonic="EUR")

    def test_base_currency_commodity_no_book(self, book_basic):
        cdty = Commodity(namespace="AMEX", mnemonic="APPLE", fullname="Apple")

        with pytest.raises(GnucashException):
            cdty.base_currency

    def test_base_currency_currency(self, book_basic):
        cdty = book_basic.currencies(mnemonic="USD")

        assert cdty.base_currency.mnemonic == "EUR"


class TestCommodity_create_prices(object):
    def test_create_basicprice(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        USD = book_basic.currencies(mnemonic="USD")
        p = Price(commodity=USD, currency=EUR, date=date(2014, 2, 22), value=Decimal('0.54321'))
        book_basic.flush()

        # check price exist
        np = USD.prices.first()
        assert np is p
        assert repr(p) == "Price<2014-02-22 : 0.54321 EUR/USD>"

        p2 = Price(commodity=USD, currency=EUR, date=date(2014, 2, 21), value=Decimal('0.12345'))
        book_basic.flush()
        assert p.value + p2.value == Decimal("0.66666")
        assert len(USD.prices.all()) == 2

    def test_create_duplicateprice(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        USD = book_basic.currencies(mnemonic="USD")
        p = Price(commodity=USD, currency=EUR, date=date(2014, 2, 22), value=Decimal('0.54321'))
        p1 = Price(commodity=USD, currency=EUR, date=date(2014, 2, 22), value=Decimal('0.12345'))

        book_basic.flush()
        assert USD.prices.filter_by(value=Decimal('0')).all() == []
        assert USD.prices.filter_by(value=Decimal('0.12345')).one() == p1

        with pytest.raises(ValueError):
            book_basic.validate()

    def test_create_duplicateprice_different_source(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        USD = book_basic.currencies(mnemonic="USD")
        p = Price(commodity=USD, currency=EUR, date=date(2014, 2, 22), value=Decimal('0.54321'), source="user:price")
        p1 = Price(commodity=USD, currency=EUR, date=date(2014, 2, 22), value=Decimal('0.12345'), source="other:price")

        book_basic.flush()
        assert USD.prices.filter_by(value=Decimal('0')).all() == []
        assert USD.prices.filter_by(value=Decimal('0.12345')).one() == p1

        # validation should work as different sources
        book_basic.validate()

    @needweb
    def test_update_currency_prices(self, book_basic):
        if not is_inmemory_sqlite(book_basic):
            print("skipping test for {}".format(book_basic))
            return

        EUR = book_basic.default_currency
        with pytest.raises(GncPriceError):
            EUR.update_prices()

        USD = book_basic.currencies(mnemonic="USD")
        USD.update_prices()

        assert len(list(USD.prices)) < 7
        assert (USD.prices.first() is None) or (USD.prices.first().commodity is USD)

        CAD = book_basic.currencies(mnemonic="CAD")
        CAD.update_prices()
        assert len(list(CAD.prices)) < 7
        assert (CAD.prices.first() is None) or (CAD.prices.first().commodity is CAD)

        # redo update prices which should not bring new prices
        l = len(list(USD.prices))
        USD.update_prices()
        assert len(list(USD.prices)) == l
        book_basic.validate()
        assert len(book_basic.prices) < 14

    @needweb
    def test_update_stock_prices(self, book_basic):
        if not is_inmemory_sqlite(book_basic):
            print("skipping test for {}".format(book_basic))
            return
        cdty = Commodity(mnemonic="AAPL", namespace="NASDAQ", fullname="Apple", book=book_basic)
        # cdty["quoted_currency"] = "USD"
        # assert cdty.get("quoted_currency") == "USD"

        cdty.update_prices()
        book_basic.flush()
        L = len(list(cdty.prices))

        assert L < 7

        cdty.update_prices()
        book_basic.flush()
        assert len(list(cdty.prices)) == L

        book_basic.validate()

    def test_price_update_on_commodity_no_book(self, book_basic):
        cdty = Commodity(namespace="AMEX", mnemonic="APPLE", fullname="Apple")

        with pytest.raises(GncPriceError):
            cdty.update_prices()



#!/usr/bin/env python
import sqlite3

import click

import piecash
from piecash.scripts.cli import cli


@cli.command()
@click.argument('book', type=click.Path(exists=True))
@click.option('--output', type=click.File('w'), default=None,
              help="File to which to dump the sql schema (default=stdout)")
def sql_dump(book, output):
    """Dump SQL schema of the gnucash sqlite book

    """

    if output is None:
        output = open(book + ".sql", "w", encoding="UTF-8")

    con = sqlite3.connect(book)
    for line in con.iterdump():
        output.write('%s\n' % line.replace("VARCHAR(", "TEXT("))

    output.close()


@cli.command()
@click.argument('book', type=click.Path(exists=False))
def sql_create(book):
    """Create an empty book with gnucash

    """
    with piecash.create_book(book, overwrite=True,keep_foreign_keys=False) as b:
        b.save()



import locale

from decimal import Decimal
from sqlalchemy import Column, VARCHAR, INTEGER, cast, Float
from sqlalchemy.ext.hybrid import hybrid_property

from .sa_extra import DeclarativeBase, _Date


class GnucashException(Exception):
    pass


class GncNoActiveSession(GnucashException):
    pass


class GncValidationError(GnucashException):
    pass


class GncImbalanceError(GncValidationError):
    pass


class GncConversionError(GnucashException):
    pass


class Recurrence(DeclarativeBase):
    """
    Recurrence information for scheduled transactions

    Attributes:
        obj_guid (str): link to the parent ScheduledTransaction record.
        recurrence_mult (int): Multiplier for the period type. Describes how many times
            the period repeats for the next occurrence.
        recurrence_period_type (str): type or recurrence (monthly, daily).
        recurrence_period_start (date): the date the recurrence starts.
        recurrence_weekend_adjust (str): adjustment to be made if the next occurrence
            falls on weekend / non-working day.
    """

    __tablename__ = 'recurrences'

    __table_args__ = {'sqlite_autoincrement': True}

    # column definitions
    id = Column('id', INTEGER(), primary_key=True, nullable=False, autoincrement=True)
    obj_guid = Column('obj_guid', VARCHAR(length=32), nullable=False)
    recurrence_mult = Column('recurrence_mult', INTEGER(), nullable=False)
    recurrence_period_type = Column('recurrence_period_type', VARCHAR(length=2048), nullable=False)
    recurrence_period_start = Column('recurrence_period_start', _Date(), nullable=False)
    recurrence_weekend_adjust = Column('recurrence_weekend_adjust', VARCHAR(length=2048),
                                       nullable=False)

    # relation definitions
    # added from the DeclarativeBaseGUID object (as linked from different objects like the slots)
    def __str__(self):
        return "{}*{} from {} [{}]".format(self.recurrence_period_type, self.recurrence_mult,
                                            self.recurrence_period_start,
                                            self.recurrence_weekend_adjust)


MAX_NUMBER = 2 ** 63 - 1


def hybrid_property_gncnumeric(num_col, denom_col):
    """Return an hybrid_property handling a Decimal represented by a numerator and a
    denominator column.
    It assumes the python field related to the sqlcolumn is named as _sqlcolumn.

    :type num_col: sqlalchemy.sql.schema.Column
    :type denom_col: sqlalchemy.sql.schema.Column
    :return: sqlalchemy.ext.hybrid.hybrid_property
    """
    num_name, denom_name = "_{}".format(num_col.name), "_{}".format(denom_col.name)
    name = num_col.name.split("_")[0]

    def fset(self, d):
        if d is None:
            num, denom = None, None
        else:
            if isinstance(d, tuple):
                d = Decimal(d[0]) / d[1]
            elif isinstance(d, (int, int, str)):
                d = Decimal(d)
            elif isinstance(d, float):
                raise TypeError(("Received a floating-point number {} where a decimal is expected. " +
                                 "Use a Decimal, str, or int instead").format(d))
            elif not isinstance(d, Decimal):
                raise TypeError(("Received an unknown type {} where a decimal is expected. " +
                                 "Use a Decimal, str, or int instead").format(type(d).__name__))

            sign, digits, exp = d.as_tuple()
            denom = 10 ** max(-exp, 0)

            denom_basis = getattr(self, "{}_basis".format(denom_name), None)
            if denom_basis is not None:
                denom = denom_basis

            num = int(d * denom)
            if not ((-MAX_NUMBER < num < MAX_NUMBER) and (-MAX_NUMBER < denom < MAX_NUMBER)):
                raise ValueError(("The amount '{}' cannot be represented in GnuCash. " +
                                  "Either it is too large or it has too many decimals").format(d))

        setattr(self, num_name, num)
        setattr(self, denom_name, denom)

    def fget(self):
        num, denom = getattr(self, num_name), getattr(self, denom_name)
        if num is None:
            return
        else:
            return Decimal(num) / denom

    def expr(cls):
        # todo: cast into Decimal for postgres and for sqlite (for the latter, use sqlite3.register_converter ?)
        return (cast(num_col, Float) / denom_col).label(name)

    return hybrid_property(
        fget=fget,
        fset=fset,
        expr=expr,
    )


class CallableList(list):
    """
    A simple class (inherited from list) allowing to retrieve a given list element with a filter on an attribute.

    It can be used as the collection_class of a sqlalchemy relationship or to wrap any list (see examples
    in :class:`piecash.core.session.GncSession`)
    """
    fallback = None

    def __init__(self, *args):
        list.__init__(self, *args)

    def __call__(self, **kwargs):
        """
        Return the first element of the list that has attributes matching the kwargs dict. The `get` method is
        an alias for this method.

        To be used as::

            l(mnemonic="EUR", namespace="CURRENCY")
        """
        for obj in self:
            for k, v in kwargs.items():
                if getattr(obj, k) != v:
                    break
            else:
                return obj
        else:
            if self.fallback:
                return self.fallback(**kwargs)
            else:
                raise KeyError("Could not find object with {} in {}".format(kwargs, self))

    get = __call__


def get_system_currency_mnemonic():
    """Returns the mnemonic of the locale currency (and EUR if not defined).

    At the target, it could also look in Gnucash configuration/registry to see if the user
    has chosen another default currency.
    """

    if locale.getlocale() == (None, None):
        locale.setlocale(locale.LC_ALL, '')

    mnemonic = locale.localeconv()['int_curr_symbol'].strip() or "EUR"

    return mnemonic



# coding=utf-8
from collections import namedtuple
from xml.etree import ElementTree

ISO_type = namedtuple("ISO_type", "country	currency	mnemonic	cusip	fraction".split("\t"))

# https://www.currency-iso.org/dam/downloads/lists/list_one.xml
ISO_currencies_XML = """<ISO_4217 Pblshd="2018-08-29">
<CcyTbl>
<CcyNtry>
<CtryNm>AFGHANISTAN</CtryNm>
<CcyNm>Afghani</CcyNm>
<Ccy>AFN</Ccy>
<CcyNbr>971</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ÅLAND ISLANDS</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ALBANIA</CtryNm>
<CcyNm>Lek</CcyNm>
<Ccy>ALL</Ccy>
<CcyNbr>008</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ALGERIA</CtryNm>
<CcyNm>Algerian Dinar</CcyNm>
<Ccy>DZD</Ccy>
<CcyNbr>012</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>AMERICAN SAMOA</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ANDORRA</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ANGOLA</CtryNm>
<CcyNm>Kwanza</CcyNm>
<Ccy>AOA</Ccy>
<CcyNbr>973</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ANGUILLA</CtryNm>
<CcyNm>East Caribbean Dollar</CcyNm>
<Ccy>XCD</Ccy>
<CcyNbr>951</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ANTARCTICA</CtryNm>
<CcyNm>No universal currency</CcyNm>
</CcyNtry>
<CcyNtry>
<CtryNm>ANTIGUA AND BARBUDA</CtryNm>
<CcyNm>East Caribbean Dollar</CcyNm>
<Ccy>XCD</Ccy>
<CcyNbr>951</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ARGENTINA</CtryNm>
<CcyNm>Argentine Peso</CcyNm>
<Ccy>ARS</Ccy>
<CcyNbr>032</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ARMENIA</CtryNm>
<CcyNm>Armenian Dram</CcyNm>
<Ccy>AMD</Ccy>
<CcyNbr>051</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ARUBA</CtryNm>
<CcyNm>Aruban Florin</CcyNm>
<Ccy>AWG</Ccy>
<CcyNbr>533</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>AUSTRALIA</CtryNm>
<CcyNm>Australian Dollar</CcyNm>
<Ccy>AUD</Ccy>
<CcyNbr>036</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>AUSTRIA</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>AZERBAIJAN</CtryNm>
<CcyNm>Azerbaijan Manat</CcyNm>
<Ccy>AZN</Ccy>
<CcyNbr>944</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BAHAMAS (THE)</CtryNm>
<CcyNm>Bahamian Dollar</CcyNm>
<Ccy>BSD</Ccy>
<CcyNbr>044</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BAHRAIN</CtryNm>
<CcyNm>Bahraini Dinar</CcyNm>
<Ccy>BHD</Ccy>
<CcyNbr>048</CcyNbr>
<CcyMnrUnts>3</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BANGLADESH</CtryNm>
<CcyNm>Taka</CcyNm>
<Ccy>BDT</Ccy>
<CcyNbr>050</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BARBADOS</CtryNm>
<CcyNm>Barbados Dollar</CcyNm>
<Ccy>BBD</Ccy>
<CcyNbr>052</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BELARUS</CtryNm>
<CcyNm>Belarusian Ruble</CcyNm>
<Ccy>BYN</Ccy>
<CcyNbr>933</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BELGIUM</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BELIZE</CtryNm>
<CcyNm>Belize Dollar</CcyNm>
<Ccy>BZD</Ccy>
<CcyNbr>084</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BENIN</CtryNm>
<CcyNm>CFA Franc BCEAO</CcyNm>
<Ccy>XOF</Ccy>
<CcyNbr>952</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BERMUDA</CtryNm>
<CcyNm>Bermudian Dollar</CcyNm>
<Ccy>BMD</Ccy>
<CcyNbr>060</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BHUTAN</CtryNm>
<CcyNm>Indian Rupee</CcyNm>
<Ccy>INR</Ccy>
<CcyNbr>356</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BHUTAN</CtryNm>
<CcyNm>Ngultrum</CcyNm>
<Ccy>BTN</Ccy>
<CcyNbr>064</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BOLIVIA (PLURINATIONAL STATE OF)</CtryNm>
<CcyNm>Boliviano</CcyNm>
<Ccy>BOB</Ccy>
<CcyNbr>068</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BOLIVIA (PLURINATIONAL STATE OF)</CtryNm>
<CcyNm IsFund="true">Mvdol</CcyNm>
<Ccy>BOV</Ccy>
<CcyNbr>984</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BONAIRE, SINT EUSTATIUS AND SABA</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BOSNIA AND HERZEGOVINA</CtryNm>
<CcyNm>Convertible Mark</CcyNm>
<Ccy>BAM</Ccy>
<CcyNbr>977</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BOTSWANA</CtryNm>
<CcyNm>Pula</CcyNm>
<Ccy>BWP</Ccy>
<CcyNbr>072</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BOUVET ISLAND</CtryNm>
<CcyNm>Norwegian Krone</CcyNm>
<Ccy>NOK</Ccy>
<CcyNbr>578</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BRAZIL</CtryNm>
<CcyNm>Brazilian Real</CcyNm>
<Ccy>BRL</Ccy>
<CcyNbr>986</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BRITISH INDIAN OCEAN TERRITORY (THE)</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BRUNEI DARUSSALAM</CtryNm>
<CcyNm>Brunei Dollar</CcyNm>
<Ccy>BND</Ccy>
<CcyNbr>096</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BULGARIA</CtryNm>
<CcyNm>Bulgarian Lev</CcyNm>
<Ccy>BGN</Ccy>
<CcyNbr>975</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BURKINA FASO</CtryNm>
<CcyNm>CFA Franc BCEAO</CcyNm>
<Ccy>XOF</Ccy>
<CcyNbr>952</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>BURUNDI</CtryNm>
<CcyNm>Burundi Franc</CcyNm>
<Ccy>BIF</Ccy>
<CcyNbr>108</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CABO VERDE</CtryNm>
<CcyNm>Cabo Verde Escudo</CcyNm>
<Ccy>CVE</Ccy>
<CcyNbr>132</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CAMBODIA</CtryNm>
<CcyNm>Riel</CcyNm>
<Ccy>KHR</Ccy>
<CcyNbr>116</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CAMEROON</CtryNm>
<CcyNm>CFA Franc BEAC</CcyNm>
<Ccy>XAF</Ccy>
<CcyNbr>950</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CANADA</CtryNm>
<CcyNm>Canadian Dollar</CcyNm>
<Ccy>CAD</Ccy>
<CcyNbr>124</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CAYMAN ISLANDS (THE)</CtryNm>
<CcyNm>Cayman Islands Dollar</CcyNm>
<Ccy>KYD</Ccy>
<CcyNbr>136</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CENTRAL AFRICAN REPUBLIC (THE)</CtryNm>
<CcyNm>CFA Franc BEAC</CcyNm>
<Ccy>XAF</Ccy>
<CcyNbr>950</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CHAD</CtryNm>
<CcyNm>CFA Franc BEAC</CcyNm>
<Ccy>XAF</Ccy>
<CcyNbr>950</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CHILE</CtryNm>
<CcyNm>Chilean Peso</CcyNm>
<Ccy>CLP</Ccy>
<CcyNbr>152</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CHILE</CtryNm>
<CcyNm IsFund="true">Unidad de Fomento</CcyNm>
<Ccy>CLF</Ccy>
<CcyNbr>990</CcyNbr>
<CcyMnrUnts>4</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CHINA</CtryNm>
<CcyNm>Yuan Renminbi</CcyNm>
<Ccy>CNY</Ccy>
<CcyNbr>156</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CHRISTMAS ISLAND</CtryNm>
<CcyNm>Australian Dollar</CcyNm>
<Ccy>AUD</Ccy>
<CcyNbr>036</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>COCOS (KEELING) ISLANDS (THE)</CtryNm>
<CcyNm>Australian Dollar</CcyNm>
<Ccy>AUD</Ccy>
<CcyNbr>036</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>COLOMBIA</CtryNm>
<CcyNm>Colombian Peso</CcyNm>
<Ccy>COP</Ccy>
<CcyNbr>170</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>COLOMBIA</CtryNm>
<CcyNm IsFund="true">Unidad de Valor Real</CcyNm>
<Ccy>COU</Ccy>
<CcyNbr>970</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>COMOROS (THE)</CtryNm>
<CcyNm>Comorian Franc </CcyNm>
<Ccy>KMF</Ccy>
<CcyNbr>174</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CONGO (THE DEMOCRATIC REPUBLIC OF THE)</CtryNm>
<CcyNm>Congolese Franc</CcyNm>
<Ccy>CDF</Ccy>
<CcyNbr>976</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CONGO (THE)</CtryNm>
<CcyNm>CFA Franc BEAC</CcyNm>
<Ccy>XAF</Ccy>
<CcyNbr>950</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>COOK ISLANDS (THE)</CtryNm>
<CcyNm>New Zealand Dollar</CcyNm>
<Ccy>NZD</Ccy>
<CcyNbr>554</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>COSTA RICA</CtryNm>
<CcyNm>Costa Rican Colon</CcyNm>
<Ccy>CRC</Ccy>
<CcyNbr>188</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CÔTE D'IVOIRE</CtryNm>
<CcyNm>CFA Franc BCEAO</CcyNm>
<Ccy>XOF</Ccy>
<CcyNbr>952</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CROATIA</CtryNm>
<CcyNm>Kuna</CcyNm>
<Ccy>HRK</Ccy>
<CcyNbr>191</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CUBA</CtryNm>
<CcyNm>Cuban Peso</CcyNm>
<Ccy>CUP</Ccy>
<CcyNbr>192</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CUBA</CtryNm>
<CcyNm>Peso Convertible</CcyNm>
<Ccy>CUC</Ccy>
<CcyNbr>931</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CURAÇAO</CtryNm>
<CcyNm>Netherlands Antillean Guilder</CcyNm>
<Ccy>ANG</Ccy>
<CcyNbr>532</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CYPRUS</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>CZECHIA</CtryNm>
<CcyNm>Czech Koruna</CcyNm>
<Ccy>CZK</Ccy>
<CcyNbr>203</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>DENMARK</CtryNm>
<CcyNm>Danish Krone</CcyNm>
<Ccy>DKK</Ccy>
<CcyNbr>208</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>DJIBOUTI</CtryNm>
<CcyNm>Djibouti Franc</CcyNm>
<Ccy>DJF</Ccy>
<CcyNbr>262</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>DOMINICA</CtryNm>
<CcyNm>East Caribbean Dollar</CcyNm>
<Ccy>XCD</Ccy>
<CcyNbr>951</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>DOMINICAN REPUBLIC (THE)</CtryNm>
<CcyNm>Dominican Peso</CcyNm>
<Ccy>DOP</Ccy>
<CcyNbr>214</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ECUADOR</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>EGYPT</CtryNm>
<CcyNm>Egyptian Pound</CcyNm>
<Ccy>EGP</Ccy>
<CcyNbr>818</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>EL SALVADOR</CtryNm>
<CcyNm>El Salvador Colon</CcyNm>
<Ccy>SVC</Ccy>
<CcyNbr>222</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>EL SALVADOR</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>EQUATORIAL GUINEA</CtryNm>
<CcyNm>CFA Franc BEAC</CcyNm>
<Ccy>XAF</Ccy>
<CcyNbr>950</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ERITREA</CtryNm>
<CcyNm>Nakfa</CcyNm>
<Ccy>ERN</Ccy>
<CcyNbr>232</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ESTONIA</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ESWATINI</CtryNm>
<CcyNm>Lilangeni</CcyNm>
<Ccy>SZL</Ccy>
<CcyNbr>748</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ETHIOPIA</CtryNm>
<CcyNm>Ethiopian Birr</CcyNm>
<Ccy>ETB</Ccy>
<CcyNbr>230</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>EUROPEAN UNION</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>FALKLAND ISLANDS (THE) [MALVINAS]</CtryNm>
<CcyNm>Falkland Islands Pound</CcyNm>
<Ccy>FKP</Ccy>
<CcyNbr>238</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>FAROE ISLANDS (THE)</CtryNm>
<CcyNm>Danish Krone</CcyNm>
<Ccy>DKK</Ccy>
<CcyNbr>208</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>FIJI</CtryNm>
<CcyNm>Fiji Dollar</CcyNm>
<Ccy>FJD</Ccy>
<CcyNbr>242</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>FINLAND</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>FRANCE</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>FRENCH GUIANA</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>FRENCH POLYNESIA</CtryNm>
<CcyNm>CFP Franc</CcyNm>
<Ccy>XPF</Ccy>
<CcyNbr>953</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>FRENCH SOUTHERN TERRITORIES (THE)</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GABON</CtryNm>
<CcyNm>CFA Franc BEAC</CcyNm>
<Ccy>XAF</Ccy>
<CcyNbr>950</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GAMBIA (THE)</CtryNm>
<CcyNm>Dalasi</CcyNm>
<Ccy>GMD</Ccy>
<CcyNbr>270</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GEORGIA</CtryNm>
<CcyNm>Lari</CcyNm>
<Ccy>GEL</Ccy>
<CcyNbr>981</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GERMANY</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GHANA</CtryNm>
<CcyNm>Ghana Cedi</CcyNm>
<Ccy>GHS</Ccy>
<CcyNbr>936</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GIBRALTAR</CtryNm>
<CcyNm>Gibraltar Pound</CcyNm>
<Ccy>GIP</Ccy>
<CcyNbr>292</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GREECE</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GREENLAND</CtryNm>
<CcyNm>Danish Krone</CcyNm>
<Ccy>DKK</Ccy>
<CcyNbr>208</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GRENADA</CtryNm>
<CcyNm>East Caribbean Dollar</CcyNm>
<Ccy>XCD</Ccy>
<CcyNbr>951</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GUADELOUPE</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GUAM</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GUATEMALA</CtryNm>
<CcyNm>Quetzal</CcyNm>
<Ccy>GTQ</Ccy>
<CcyNbr>320</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GUERNSEY</CtryNm>
<CcyNm>Pound Sterling</CcyNm>
<Ccy>GBP</Ccy>
<CcyNbr>826</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GUINEA</CtryNm>
<CcyNm>Guinean Franc</CcyNm>
<Ccy>GNF</Ccy>
<CcyNbr>324</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GUINEA-BISSAU</CtryNm>
<CcyNm>CFA Franc BCEAO</CcyNm>
<Ccy>XOF</Ccy>
<CcyNbr>952</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>GUYANA</CtryNm>
<CcyNm>Guyana Dollar</CcyNm>
<Ccy>GYD</Ccy>
<CcyNbr>328</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>HAITI</CtryNm>
<CcyNm>Gourde</CcyNm>
<Ccy>HTG</Ccy>
<CcyNbr>332</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>HAITI</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>HEARD ISLAND AND McDONALD ISLANDS</CtryNm>
<CcyNm>Australian Dollar</CcyNm>
<Ccy>AUD</Ccy>
<CcyNbr>036</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>HOLY SEE (THE)</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>HONDURAS</CtryNm>
<CcyNm>Lempira</CcyNm>
<Ccy>HNL</Ccy>
<CcyNbr>340</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>HONG KONG</CtryNm>
<CcyNm>Hong Kong Dollar</CcyNm>
<Ccy>HKD</Ccy>
<CcyNbr>344</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>HUNGARY</CtryNm>
<CcyNm>Forint</CcyNm>
<Ccy>HUF</Ccy>
<CcyNbr>348</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ICELAND</CtryNm>
<CcyNm>Iceland Krona</CcyNm>
<Ccy>ISK</Ccy>
<CcyNbr>352</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>INDIA</CtryNm>
<CcyNm>Indian Rupee</CcyNm>
<Ccy>INR</Ccy>
<CcyNbr>356</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>INDONESIA</CtryNm>
<CcyNm>Rupiah</CcyNm>
<Ccy>IDR</Ccy>
<CcyNbr>360</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>INTERNATIONAL MONETARY FUND (IMF) </CtryNm>
<CcyNm>SDR (Special Drawing Right)</CcyNm>
<Ccy>XDR</Ccy>
<CcyNbr>960</CcyNbr>
<CcyMnrUnts>N.A.</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>IRAN (ISLAMIC REPUBLIC OF)</CtryNm>
<CcyNm>Iranian Rial</CcyNm>
<Ccy>IRR</Ccy>
<CcyNbr>364</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>IRAQ</CtryNm>
<CcyNm>Iraqi Dinar</CcyNm>
<Ccy>IQD</Ccy>
<CcyNbr>368</CcyNbr>
<CcyMnrUnts>3</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>IRELAND</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ISLE OF MAN</CtryNm>
<CcyNm>Pound Sterling</CcyNm>
<Ccy>GBP</Ccy>
<CcyNbr>826</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ISRAEL</CtryNm>
<CcyNm>New Israeli Sheqel</CcyNm>
<Ccy>ILS</Ccy>
<CcyNbr>376</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ITALY</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>JAMAICA</CtryNm>
<CcyNm>Jamaican Dollar</CcyNm>
<Ccy>JMD</Ccy>
<CcyNbr>388</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>JAPAN</CtryNm>
<CcyNm>Yen</CcyNm>
<Ccy>JPY</Ccy>
<CcyNbr>392</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>JERSEY</CtryNm>
<CcyNm>Pound Sterling</CcyNm>
<Ccy>GBP</Ccy>
<CcyNbr>826</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>JORDAN</CtryNm>
<CcyNm>Jordanian Dinar</CcyNm>
<Ccy>JOD</Ccy>
<CcyNbr>400</CcyNbr>
<CcyMnrUnts>3</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>KAZAKHSTAN</CtryNm>
<CcyNm>Tenge</CcyNm>
<Ccy>KZT</Ccy>
<CcyNbr>398</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>KENYA</CtryNm>
<CcyNm>Kenyan Shilling</CcyNm>
<Ccy>KES</Ccy>
<CcyNbr>404</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>KIRIBATI</CtryNm>
<CcyNm>Australian Dollar</CcyNm>
<Ccy>AUD</Ccy>
<CcyNbr>036</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>KOREA (THE DEMOCRATIC PEOPLE’S REPUBLIC OF)</CtryNm>
<CcyNm>North Korean Won</CcyNm>
<Ccy>KPW</Ccy>
<CcyNbr>408</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>KOREA (THE REPUBLIC OF)</CtryNm>
<CcyNm>Won</CcyNm>
<Ccy>KRW</Ccy>
<CcyNbr>410</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>KUWAIT</CtryNm>
<CcyNm>Kuwaiti Dinar</CcyNm>
<Ccy>KWD</Ccy>
<CcyNbr>414</CcyNbr>
<CcyMnrUnts>3</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>KYRGYZSTAN</CtryNm>
<CcyNm>Som</CcyNm>
<Ccy>KGS</Ccy>
<CcyNbr>417</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>LAO PEOPLE’S DEMOCRATIC REPUBLIC (THE)</CtryNm>
<CcyNm>Lao Kip</CcyNm>
<Ccy>LAK</Ccy>
<CcyNbr>418</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>LATVIA</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>LEBANON</CtryNm>
<CcyNm>Lebanese Pound</CcyNm>
<Ccy>LBP</Ccy>
<CcyNbr>422</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>LESOTHO</CtryNm>
<CcyNm>Loti</CcyNm>
<Ccy>LSL</Ccy>
<CcyNbr>426</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>LESOTHO</CtryNm>
<CcyNm>Rand</CcyNm>
<Ccy>ZAR</Ccy>
<CcyNbr>710</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>LIBERIA</CtryNm>
<CcyNm>Liberian Dollar</CcyNm>
<Ccy>LRD</Ccy>
<CcyNbr>430</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>LIBYA</CtryNm>
<CcyNm>Libyan Dinar</CcyNm>
<Ccy>LYD</Ccy>
<CcyNbr>434</CcyNbr>
<CcyMnrUnts>3</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>LIECHTENSTEIN</CtryNm>
<CcyNm>Swiss Franc</CcyNm>
<Ccy>CHF</Ccy>
<CcyNbr>756</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>LITHUANIA</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>LUXEMBOURG</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MACAO</CtryNm>
<CcyNm>Pataca</CcyNm>
<Ccy>MOP</Ccy>
<CcyNbr>446</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NORTH MACEDONIA</CtryNm>
<CcyNm>Denar</CcyNm>
<Ccy>MKD</Ccy>
<CcyNbr>807</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MADAGASCAR</CtryNm>
<CcyNm>Malagasy Ariary</CcyNm>
<Ccy>MGA</Ccy>
<CcyNbr>969</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MALAWI</CtryNm>
<CcyNm>Malawi Kwacha</CcyNm>
<Ccy>MWK</Ccy>
<CcyNbr>454</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MALAYSIA</CtryNm>
<CcyNm>Malaysian Ringgit</CcyNm>
<Ccy>MYR</Ccy>
<CcyNbr>458</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MALDIVES</CtryNm>
<CcyNm>Rufiyaa</CcyNm>
<Ccy>MVR</Ccy>
<CcyNbr>462</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MALI</CtryNm>
<CcyNm>CFA Franc BCEAO</CcyNm>
<Ccy>XOF</Ccy>
<CcyNbr>952</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MALTA</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MARSHALL ISLANDS (THE)</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MARTINIQUE</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MAURITANIA</CtryNm>
<CcyNm>Ouguiya</CcyNm>
<Ccy>MRU</Ccy>
<CcyNbr>929</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MAURITIUS</CtryNm>
<CcyNm>Mauritius Rupee</CcyNm>
<Ccy>MUR</Ccy>
<CcyNbr>480</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MAYOTTE</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MEMBER COUNTRIES OF THE AFRICAN DEVELOPMENT BANK GROUP</CtryNm>
<CcyNm>ADB Unit of Account</CcyNm>
<Ccy>XUA</Ccy>
<CcyNbr>965</CcyNbr>
<CcyMnrUnts>N.A.</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MEXICO</CtryNm>
<CcyNm>Mexican Peso</CcyNm>
<Ccy>MXN</Ccy>
<CcyNbr>484</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MEXICO</CtryNm>
<CcyNm IsFund="true">Mexican Unidad de Inversion (UDI)</CcyNm>
<Ccy>MXV</Ccy>
<CcyNbr>979</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MICRONESIA (FEDERATED STATES OF)</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MOLDOVA (THE REPUBLIC OF)</CtryNm>
<CcyNm>Moldovan Leu</CcyNm>
<Ccy>MDL</Ccy>
<CcyNbr>498</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MONACO</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MONGOLIA</CtryNm>
<CcyNm>Tugrik</CcyNm>
<Ccy>MNT</Ccy>
<CcyNbr>496</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MONTENEGRO</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MONTSERRAT</CtryNm>
<CcyNm>East Caribbean Dollar</CcyNm>
<Ccy>XCD</Ccy>
<CcyNbr>951</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MOROCCO</CtryNm>
<CcyNm>Moroccan Dirham</CcyNm>
<Ccy>MAD</Ccy>
<CcyNbr>504</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MOZAMBIQUE</CtryNm>
<CcyNm>Mozambique Metical</CcyNm>
<Ccy>MZN</Ccy>
<CcyNbr>943</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>MYANMAR</CtryNm>
<CcyNm>Kyat</CcyNm>
<Ccy>MMK</Ccy>
<CcyNbr>104</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NAMIBIA</CtryNm>
<CcyNm>Namibia Dollar</CcyNm>
<Ccy>NAD</Ccy>
<CcyNbr>516</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NAMIBIA</CtryNm>
<CcyNm>Rand</CcyNm>
<Ccy>ZAR</Ccy>
<CcyNbr>710</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NAURU</CtryNm>
<CcyNm>Australian Dollar</CcyNm>
<Ccy>AUD</Ccy>
<CcyNbr>036</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NEPAL</CtryNm>
<CcyNm>Nepalese Rupee</CcyNm>
<Ccy>NPR</Ccy>
<CcyNbr>524</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NETHERLANDS (THE)</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NEW CALEDONIA</CtryNm>
<CcyNm>CFP Franc</CcyNm>
<Ccy>XPF</Ccy>
<CcyNbr>953</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NEW ZEALAND</CtryNm>
<CcyNm>New Zealand Dollar</CcyNm>
<Ccy>NZD</Ccy>
<CcyNbr>554</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NICARAGUA</CtryNm>
<CcyNm>Cordoba Oro</CcyNm>
<Ccy>NIO</Ccy>
<CcyNbr>558</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NIGER (THE)</CtryNm>
<CcyNm>CFA Franc BCEAO</CcyNm>
<Ccy>XOF</Ccy>
<CcyNbr>952</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NIGERIA</CtryNm>
<CcyNm>Naira</CcyNm>
<Ccy>NGN</Ccy>
<CcyNbr>566</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NIUE</CtryNm>
<CcyNm>New Zealand Dollar</CcyNm>
<Ccy>NZD</Ccy>
<CcyNbr>554</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NORFOLK ISLAND</CtryNm>
<CcyNm>Australian Dollar</CcyNm>
<Ccy>AUD</Ccy>
<CcyNbr>036</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NORTHERN MARIANA ISLANDS (THE)</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>NORWAY</CtryNm>
<CcyNm>Norwegian Krone</CcyNm>
<Ccy>NOK</Ccy>
<CcyNbr>578</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>OMAN</CtryNm>
<CcyNm>Rial Omani</CcyNm>
<Ccy>OMR</Ccy>
<CcyNbr>512</CcyNbr>
<CcyMnrUnts>3</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>PAKISTAN</CtryNm>
<CcyNm>Pakistan Rupee</CcyNm>
<Ccy>PKR</Ccy>
<CcyNbr>586</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>PALAU</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>PALESTINE, STATE OF</CtryNm>
<CcyNm>No universal currency</CcyNm>
</CcyNtry>
<CcyNtry>
<CtryNm>PANAMA</CtryNm>
<CcyNm>Balboa</CcyNm>
<Ccy>PAB</Ccy>
<CcyNbr>590</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>PANAMA</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>PAPUA NEW GUINEA</CtryNm>
<CcyNm>Kina</CcyNm>
<Ccy>PGK</Ccy>
<CcyNbr>598</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>PARAGUAY</CtryNm>
<CcyNm>Guarani</CcyNm>
<Ccy>PYG</Ccy>
<CcyNbr>600</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>PERU</CtryNm>
<CcyNm>Sol</CcyNm>
<Ccy>PEN</Ccy>
<CcyNbr>604</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>PHILIPPINES (THE)</CtryNm>
<CcyNm>Philippine Peso</CcyNm>
<Ccy>PHP</Ccy>
<CcyNbr>608</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>PITCAIRN</CtryNm>
<CcyNm>New Zealand Dollar</CcyNm>
<Ccy>NZD</Ccy>
<CcyNbr>554</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>POLAND</CtryNm>
<CcyNm>Zloty</CcyNm>
<Ccy>PLN</Ccy>
<CcyNbr>985</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>PORTUGAL</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>PUERTO RICO</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>QATAR</CtryNm>
<CcyNm>Qatari Rial</CcyNm>
<Ccy>QAR</Ccy>
<CcyNbr>634</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>RÉUNION</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ROMANIA</CtryNm>
<CcyNm>Romanian Leu</CcyNm>
<Ccy>RON</Ccy>
<CcyNbr>946</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>RUSSIAN FEDERATION (THE)</CtryNm>
<CcyNm>Russian Ruble</CcyNm>
<Ccy>RUB</Ccy>
<CcyNbr>643</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>RWANDA</CtryNm>
<CcyNm>Rwanda Franc</CcyNm>
<Ccy>RWF</Ccy>
<CcyNbr>646</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SAINT BARTHÉLEMY</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SAINT HELENA, ASCENSION AND TRISTAN DA CUNHA</CtryNm>
<CcyNm>Saint Helena Pound</CcyNm>
<Ccy>SHP</Ccy>
<CcyNbr>654</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SAINT KITTS AND NEVIS</CtryNm>
<CcyNm>East Caribbean Dollar</CcyNm>
<Ccy>XCD</Ccy>
<CcyNbr>951</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SAINT LUCIA</CtryNm>
<CcyNm>East Caribbean Dollar</CcyNm>
<Ccy>XCD</Ccy>
<CcyNbr>951</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SAINT MARTIN (FRENCH PART)</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SAINT PIERRE AND MIQUELON</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SAINT VINCENT AND THE GRENADINES</CtryNm>
<CcyNm>East Caribbean Dollar</CcyNm>
<Ccy>XCD</Ccy>
<CcyNbr>951</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SAMOA</CtryNm>
<CcyNm>Tala</CcyNm>
<Ccy>WST</Ccy>
<CcyNbr>882</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SAN MARINO</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SAO TOME AND PRINCIPE</CtryNm>
<CcyNm>Dobra</CcyNm>
<Ccy>STN</Ccy>
<CcyNbr>930</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SAUDI ARABIA</CtryNm>
<CcyNm>Saudi Riyal</CcyNm>
<Ccy>SAR</Ccy>
<CcyNbr>682</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SENEGAL</CtryNm>
<CcyNm>CFA Franc BCEAO</CcyNm>
<Ccy>XOF</Ccy>
<CcyNbr>952</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SERBIA</CtryNm>
<CcyNm>Serbian Dinar</CcyNm>
<Ccy>RSD</Ccy>
<CcyNbr>941</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SEYCHELLES</CtryNm>
<CcyNm>Seychelles Rupee</CcyNm>
<Ccy>SCR</Ccy>
<CcyNbr>690</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SIERRA LEONE</CtryNm>
<CcyNm>Leone</CcyNm>
<Ccy>SLL</Ccy>
<CcyNbr>694</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SINGAPORE</CtryNm>
<CcyNm>Singapore Dollar</CcyNm>
<Ccy>SGD</Ccy>
<CcyNbr>702</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SINT MAARTEN (DUTCH PART)</CtryNm>
<CcyNm>Netherlands Antillean Guilder</CcyNm>
<Ccy>ANG</Ccy>
<CcyNbr>532</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SISTEMA UNITARIO DE COMPENSACION REGIONAL DE PAGOS "SUCRE"</CtryNm>
<CcyNm>Sucre</CcyNm>
<Ccy>XSU</Ccy>
<CcyNbr>994</CcyNbr>
<CcyMnrUnts>N.A.</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SLOVAKIA</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SLOVENIA</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SOLOMON ISLANDS</CtryNm>
<CcyNm>Solomon Islands Dollar</CcyNm>
<Ccy>SBD</Ccy>
<CcyNbr>090</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SOMALIA</CtryNm>
<CcyNm>Somali Shilling</CcyNm>
<Ccy>SOS</Ccy>
<CcyNbr>706</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SOUTH AFRICA</CtryNm>
<CcyNm>Rand</CcyNm>
<Ccy>ZAR</Ccy>
<CcyNbr>710</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SOUTH GEORGIA AND THE SOUTH SANDWICH ISLANDS</CtryNm>
<CcyNm>No universal currency</CcyNm>
</CcyNtry>
<CcyNtry>
<CtryNm>SOUTH SUDAN</CtryNm>
<CcyNm>South Sudanese Pound</CcyNm>
<Ccy>SSP</Ccy>
<CcyNbr>728</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SPAIN</CtryNm>
<CcyNm>Euro</CcyNm>
<Ccy>EUR</Ccy>
<CcyNbr>978</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SRI LANKA</CtryNm>
<CcyNm>Sri Lanka Rupee</CcyNm>
<Ccy>LKR</Ccy>
<CcyNbr>144</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SUDAN (THE)</CtryNm>
<CcyNm>Sudanese Pound</CcyNm>
<Ccy>SDG</Ccy>
<CcyNbr>938</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SURINAME</CtryNm>
<CcyNm>Surinam Dollar</CcyNm>
<Ccy>SRD</Ccy>
<CcyNbr>968</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SVALBARD AND JAN MAYEN</CtryNm>
<CcyNm>Norwegian Krone</CcyNm>
<Ccy>NOK</Ccy>
<CcyNbr>578</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SWEDEN</CtryNm>
<CcyNm>Swedish Krona</CcyNm>
<Ccy>SEK</Ccy>
<CcyNbr>752</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SWITZERLAND</CtryNm>
<CcyNm>Swiss Franc</CcyNm>
<Ccy>CHF</Ccy>
<CcyNbr>756</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SWITZERLAND</CtryNm>
<CcyNm IsFund="true">WIR Euro</CcyNm>
<Ccy>CHE</Ccy>
<CcyNbr>947</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SWITZERLAND</CtryNm>
<CcyNm IsFund="true">WIR Franc</CcyNm>
<Ccy>CHW</Ccy>
<CcyNbr>948</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>SYRIAN ARAB REPUBLIC</CtryNm>
<CcyNm>Syrian Pound</CcyNm>
<Ccy>SYP</Ccy>
<CcyNbr>760</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>TAIWAN (PROVINCE OF CHINA)</CtryNm>
<CcyNm>New Taiwan Dollar</CcyNm>
<Ccy>TWD</Ccy>
<CcyNbr>901</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>TAJIKISTAN</CtryNm>
<CcyNm>Somoni</CcyNm>
<Ccy>TJS</Ccy>
<CcyNbr>972</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>TANZANIA, UNITED REPUBLIC OF</CtryNm>
<CcyNm>Tanzanian Shilling</CcyNm>
<Ccy>TZS</Ccy>
<CcyNbr>834</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>THAILAND</CtryNm>
<CcyNm>Baht</CcyNm>
<Ccy>THB</Ccy>
<CcyNbr>764</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>TIMOR-LESTE</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>TOGO</CtryNm>
<CcyNm>CFA Franc BCEAO</CcyNm>
<Ccy>XOF</Ccy>
<CcyNbr>952</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>TOKELAU</CtryNm>
<CcyNm>New Zealand Dollar</CcyNm>
<Ccy>NZD</Ccy>
<CcyNbr>554</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>TONGA</CtryNm>
<CcyNm>Pa’anga</CcyNm>
<Ccy>TOP</Ccy>
<CcyNbr>776</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>TRINIDAD AND TOBAGO</CtryNm>
<CcyNm>Trinidad and Tobago Dollar</CcyNm>
<Ccy>TTD</Ccy>
<CcyNbr>780</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>TUNISIA</CtryNm>
<CcyNm>Tunisian Dinar</CcyNm>
<Ccy>TND</Ccy>
<CcyNbr>788</CcyNbr>
<CcyMnrUnts>3</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>TURKEY</CtryNm>
<CcyNm>Turkish Lira</CcyNm>
<Ccy>TRY</Ccy>
<CcyNbr>949</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>TURKMENISTAN</CtryNm>
<CcyNm>Turkmenistan New Manat</CcyNm>
<Ccy>TMT</Ccy>
<CcyNbr>934</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>TURKS AND CAICOS ISLANDS (THE)</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>TUVALU</CtryNm>
<CcyNm>Australian Dollar</CcyNm>
<Ccy>AUD</Ccy>
<CcyNbr>036</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>UGANDA</CtryNm>
<CcyNm>Uganda Shilling</CcyNm>
<Ccy>UGX</Ccy>
<CcyNbr>800</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>UKRAINE</CtryNm>
<CcyNm>Hryvnia</CcyNm>
<Ccy>UAH</Ccy>
<CcyNbr>980</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>UNITED ARAB EMIRATES (THE)</CtryNm>
<CcyNm>UAE Dirham</CcyNm>
<Ccy>AED</Ccy>
<CcyNbr>784</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>UNITED KINGDOM OF GREAT BRITAIN AND NORTHERN IRELAND (THE)</CtryNm>
<CcyNm>Pound Sterling</CcyNm>
<Ccy>GBP</Ccy>
<CcyNbr>826</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>UNITED STATES MINOR OUTLYING ISLANDS (THE)</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>UNITED STATES OF AMERICA (THE)</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>UNITED STATES OF AMERICA (THE)</CtryNm>
<CcyNm IsFund="true">US Dollar (Next day)</CcyNm>
<Ccy>USN</Ccy>
<CcyNbr>997</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>URUGUAY</CtryNm>
<CcyNm>Peso Uruguayo</CcyNm>
<Ccy>UYU</Ccy>
<CcyNbr>858</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>URUGUAY</CtryNm>
<CcyNm IsFund="true">Uruguay Peso en Unidades Indexadas (UI)</CcyNm>
<Ccy>UYI</Ccy>
<CcyNbr>940</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>URUGUAY</CtryNm>
<CcyNm>Unidad Previsional</CcyNm>
<Ccy>UYW</Ccy>
<CcyNbr>927</CcyNbr>
<CcyMnrUnts>4</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>UZBEKISTAN</CtryNm>
<CcyNm>Uzbekistan Sum</CcyNm>
<Ccy>UZS</Ccy>
<CcyNbr>860</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>VANUATU</CtryNm>
<CcyNm>Vatu</CcyNm>
<Ccy>VUV</Ccy>
<CcyNbr>548</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>VENEZUELA (BOLIVARIAN REPUBLIC OF)</CtryNm>
<CcyNm>Bolívar Soberano</CcyNm>
<Ccy>VES</Ccy>
<CcyNbr>928</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>VIET NAM</CtryNm>
<CcyNm>Dong</CcyNm>
<Ccy>VND</Ccy>
<CcyNbr>704</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>VIRGIN ISLANDS (BRITISH)</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>VIRGIN ISLANDS (U.S.)</CtryNm>
<CcyNm>US Dollar</CcyNm>
<Ccy>USD</Ccy>
<CcyNbr>840</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>WALLIS AND FUTUNA</CtryNm>
<CcyNm>CFP Franc</CcyNm>
<Ccy>XPF</Ccy>
<CcyNbr>953</CcyNbr>
<CcyMnrUnts>0</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>WESTERN SAHARA</CtryNm>
<CcyNm>Moroccan Dirham</CcyNm>
<Ccy>MAD</Ccy>
<CcyNbr>504</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>YEMEN</CtryNm>
<CcyNm>Yemeni Rial</CcyNm>
<Ccy>YER</Ccy>
<CcyNbr>886</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ZAMBIA</CtryNm>
<CcyNm>Zambian Kwacha</CcyNm>
<Ccy>ZMW</Ccy>
<CcyNbr>967</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ZIMBABWE</CtryNm>
<CcyNm>Zimbabwe Dollar</CcyNm>
<Ccy>ZWL</Ccy>
<CcyNbr>932</CcyNbr>
<CcyMnrUnts>2</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ZZ01_Bond Markets Unit European_EURCO</CtryNm>
<CcyNm>Bond Markets Unit European Composite Unit (EURCO)</CcyNm>
<Ccy>XBA</Ccy>
<CcyNbr>955</CcyNbr>
<CcyMnrUnts>N.A.</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ZZ02_Bond Markets Unit European_EMU-6</CtryNm>
<CcyNm>Bond Markets Unit European Monetary Unit (E.M.U.-6)</CcyNm>
<Ccy>XBB</Ccy>
<CcyNbr>956</CcyNbr>
<CcyMnrUnts>N.A.</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ZZ03_Bond Markets Unit European_EUA-9</CtryNm>
<CcyNm>Bond Markets Unit European Unit of Account 9 (E.U.A.-9)</CcyNm>
<Ccy>XBC</Ccy>
<CcyNbr>957</CcyNbr>
<CcyMnrUnts>N.A.</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ZZ04_Bond Markets Unit European_EUA-17</CtryNm>
<CcyNm>Bond Markets Unit European Unit of Account 17 (E.U.A.-17)</CcyNm>
<Ccy>XBD</Ccy>
<CcyNbr>958</CcyNbr>
<CcyMnrUnts>N.A.</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ZZ06_Testing_Code</CtryNm>
<CcyNm>Codes specifically reserved for testing purposes</CcyNm>
<Ccy>XTS</Ccy>
<CcyNbr>963</CcyNbr>
<CcyMnrUnts>N.A.</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ZZ07_No_Currency</CtryNm>
<CcyNm>The codes assigned for transactions where no currency is involved</CcyNm>
<Ccy>XXX</Ccy>
<CcyNbr>999</CcyNbr>
<CcyMnrUnts>N.A.</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ZZ08_Gold</CtryNm>
<CcyNm>Gold</CcyNm>
<Ccy>XAU</Ccy>
<CcyNbr>959</CcyNbr>
<CcyMnrUnts>N.A.</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ZZ09_Palladium</CtryNm>
<CcyNm>Palladium</CcyNm>
<Ccy>XPD</Ccy>
<CcyNbr>964</CcyNbr>
<CcyMnrUnts>N.A.</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ZZ10_Platinum</CtryNm>
<CcyNm>Platinum</CcyNm>
<Ccy>XPT</Ccy>
<CcyNbr>962</CcyNbr>
<CcyMnrUnts>N.A.</CcyMnrUnts>
</CcyNtry>
<CcyNtry>
<CtryNm>ZZ11_Silver</CtryNm>
<CcyNm>Silver</CcyNm>
<Ccy>XAG</Ccy>
<CcyNbr>961</CcyNbr>
<CcyMnrUnts>N.A.</CcyMnrUnts>
</CcyNtry>
</CcyTbl>
</ISO_4217>
"""
ISO_currencies = {cur.findtext("Ccy"): ISO_type(*[e.text for e in cur])
                  for cur in ElementTree.fromstring(ISO_currencies_XML).findall(".//CcyNtry")
                  if cur.findtext("CcyMnrUnts")
                  }



# coding=utf-8
from __future__ import unicode_literals

from decimal import Decimal

# dummy line to avoid removing unused symbols
from piecash import Address, Employee, Account, Vendor, Customer
from piecash.business import Taxtable, TaxtableEntry
from test_helper import db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_basic, Person

a = db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_basic, Person


class TestBusinessPerson_create_Person(object):
    """
    Person is a parameter taking values in [Customer, Vendor, Employee]
    """

    def test_create_person_noid_nobook(self, book_basic, Person):
        EUR = book_basic.commodities(namespace="CURRENCY")

        # create detached person
        c = Person(name="John Föo", currency=EUR)
        # id should not be set
        assert c.id is None

        # flushing should not set the id as person not added to book
        book_basic.flush()
        assert c.id is None

        # adding the person to the book does not per se set the id
        book_basic.add(c)
        assert c.id == '000001'
        # but validation sets the id if still to None
        assert getattr(book_basic, Person._counter_name) == 1

        book_basic.flush()
        assert c.id == "000001"
        assert getattr(book_basic, Person._counter_name) == 1

    def test_create_person_noid_inbook(self, book_basic, Person):
        EUR = book_basic.commodities(namespace="CURRENCY")

        # create person attached to book
        c = Person(name="John Föo", currency=EUR, book=book_basic)
        # id should have already been set
        assert c.id == "000001"

        # flushing should not change the id
        book_basic.flush()
        assert c.id == "000001"

    def test_create_person_id_inbook(self, book_basic, Person):
        EUR = book_basic.commodities(namespace="CURRENCY")

        # create person attached to book with a specific id
        cust_id = "éyO903"
        c = Person(name="John Föo", currency=EUR, book=book_basic, id=cust_id)
        # id should have already been set
        assert c.id == cust_id

        # flushing should not change the id
        book_basic.flush()
        assert c.id == cust_id

    def test_create_person_id_nobook(self, book_basic, Person):
        EUR = book_basic.commodities(namespace="CURRENCY")

        # create person detached from book with a specific id
        cust_id = "éyO903"
        c = Person(name="John Föo", currency=EUR, id=cust_id)
        # id should have already been set
        assert c.id == cust_id

        # flushing should not change the id (as the person is not yet attached to book)
        book_basic.flush()
        assert c.id == cust_id

        # adding the person to the book and flushing should not change the id
        book_basic.add(c)
        assert c.id == cust_id
        book_basic.flush()
        assert c.id == cust_id

    def test_create_person_address(self, book_basic, Person):
        EUR = book_basic.commodities(namespace="CURRENCY")

        # create person detached from book with a specific id
        addr = Address(name="Héllo", addr1="kap", email="foo@example.com")
        c = Person(name="John Föo", currency=EUR, address=addr, book=book_basic)

        assert c.addr_addr1 == "kap"
        assert c.address.addr1 == "kap"

        addr.addr1 = "pok"
        c2 = Person(name="Jané Döo", currency=EUR, address=addr, book=book_basic)
        book_basic.flush()

        assert c.addr_addr1 == "kap"
        assert c.address.addr1 == "kap"
        assert c2.addr_addr1 == "pok"
        assert c2.address.addr1 == "pok"

    def test_create_person_taxtabme(self, book_basic, Person):
        if Person is Employee:
            return

        EUR = book_basic.commodities(namespace="CURRENCY")

        # create person detached from book with a specific id
        taxtable = Taxtable(name="Local tax", entries=[
            TaxtableEntry(type="percentage",
                          amount=Decimal("6.5"),
                          account=Account(name="MyAcc", parent=book_basic.root_account, commodity=EUR, type="ASSET"))
        ])
        te = TaxtableEntry(type="percentage",
                           amount=Decimal("6.5"),
                           account=Account(name="MyOtherAcc", parent=book_basic.root_account, commodity=EUR,
                                           type="ASSET"),
                           taxtable=taxtable)

        c = Person(name="John Föo", currency=EUR, taxtable=taxtable, book=book_basic)
        assert len(taxtable.entries) == 2
        assert taxtable.entries[0].account.parent == book_basic.root_account

        book_basic.flush()

        assert book_basic.taxtables == [taxtable]



# -*- coding: latin-1 -*-

from pathlib import Path

import pytest

import piecash
from test_helper import book_complex

# dummy line to avoid removing unused symbols
a = book_complex

REFERENCE = Path(__file__).parent / "references"


@pytest.mark.parametrize(
    "options",
    [dict(), dict(locale=True), dict(commodity_notes=True), dict(short_account_names=True)],
)
def test_out_write(book_complex, options):
    ledger_output = piecash.ledger(book_complex, **options)

    file_name = "file_template_full" + "".join(f".{k}_{v}" for k, v in options.items()) + ".ledger"

    # to generate the first time the expected output of the test
    (REFERENCE /  file_name).write_text(ledger_output, encoding="utf-8")

    assert ledger_output == (REFERENCE / file_name).read_text(encoding="utf-8")


def test_short_account_names_raise_error_when_duplicate_account_names(book_complex):
    # no exception
    piecash.ledger(book_complex, short_account_names=True)

    # exception as two accounts have the same short name
    book_complex.accounts[0].name = book_complex.accounts[1].name
    book_complex.flush()
    with pytest.raises(
        ValueError,
        match="You have duplicate short names in your book. "
        "You cannot use the 'short_account_names' option.",
    ):
        piecash.ledger(book_complex, short_account_names=True)
